@@ -1,4 +1,5 @@
 from gitflow.core import GitFlow
+import httplib2
 import busyflow.pivotal as pt
 import string
 from colorama import Style
@@ -7,12 +8,15 @@ init()
 import sys
 import re
 import itertools
+import json
 
 from gitflow.exceptions import (NotInitialized, GitflowError,
                                 ReleaseAlreadyAssigned, IllegalVersionFormat,
-                                StatusError, NoSuchBranchError)
+                                StatusError, NoSuchBranchError, PointMeError)
 
 
+PT_V5_ENDPOINT = 'https://www.pivotaltracker.com/services/v5'
+
 _gitflow = GitFlow()
 
 def check_version_format(version):
@@ -23,6 +27,9 @@ def check_version_format(version):
 def _get_version_matcher():
     return _gitflow._safe_get('gitflow.release.versionmatcher')
 
+def _get_token():
+    return _gitflow._safe_get('gitflow.pt.token')
+
 def _get_client():
     return pt.PivotalClient(token=_gitflow._safe_get('gitflow.pt.token'))
 
@@ -54,7 +61,23 @@ def _iter_backlog_stories():
                 yield s
 
 def _iter_stories():
-    return itertools.chain(_iter_current_stories(), _iter_backlog_stories())
+    # Load the PT include/exclude labels from git config.
+    # Use string.lower() since PT labels are case insensitive.
+    include = _gitflow.get('gitflow.pt.includelabel', None)
+    if include is not None:
+        include = include.lower()
+
+    exclude = _gitflow.get('gitflow.pt.excludelabels', None)
+    if exclude is not None:
+        exclude = exclude.lower()
+        exclude = exclude.split(',')
+
+    for s in itertools.chain(_iter_current_stories(), _iter_backlog_stories()):
+        if include and not s.is_labeled(include):
+            continue
+        if exclude and any([l for l in exclude if s.is_labeled(l)]):
+            continue
+        yield s
 
 def list_projects():
     projects = _get_client().projects.all()['projects']
@@ -111,12 +134,45 @@ class Story(object):
         return self.get_state() == 'started'
 
     def is_unstarted(self):
-        return self.get_state() == 'unstarted'
+        return self.get_state() == 'unstarted' or self.get_state() == 'unscheduled'
+
+    def set_me_as_owner(self):
+        headers = {'X-TrackerToken': _get_token()}
+
+        # Download /me
+        endpoint = PT_V5_ENDPOINT + '/me'
+        resp, content = httplib2.Http().request(
+                endpoint,
+                headers=headers)
+        if resp.status != 200:
+            raise OperationsError('failed to get Pivotal Tracker /me')
+
+        # Decode my ID
+        my_id = None
+        try:
+            my_id = json.loads(content)['id']
+        except ValueError:
+            raise OperationsError('failed to decode Pivotal Tracker /me response')
+
+        # Set my ID as the owner of this story
+        endpoint = '{0}/projects/{1}/stories/{2}' \
+                   .format(PT_V5_ENDPOINT, _get_project_id(), self.get_id())
+        req_content = json.dumps({'owned_by_id': my_id})
+        headers['Content-Type'] = 'application/json'
+        rep, rep_content = httplib2.Http().request(
+                endpoint,
+                body=req_content,
+                method='PUT',
+                headers=headers)
+        if rep.status != 200:
+            raise OperationsError('failed to set me as the story owner')
 
     #+++ Bug- & Feature-specific stuff
     def finish(self):
         assert self.is_feature() or self.is_bug()
-        if self.get_state() != 'started':
+        if self.is_finished():
+            return
+        if not self.is_started():
             raise StatusError('Feature not started: ' + str(self.get_id()))
         self.set_state('finished')
 
@@ -126,7 +182,9 @@ class Story(object):
 
     def deliver(self):
         assert self.is_feature() or self.is_bug()
-        if self.get_state() != 'finished':
+        if self.is_delivered():
+            return
+        if not self.is_finished():
             raise StatusError('Feature not finished: ' + str(self.get_id()))
         self.set_state('delivered')
 
@@ -228,6 +286,18 @@ class Release(object):
     def __init__(self, version, _skip_story_download=False):
         check_version_format(version)
         self._version = version
+
+        # Load the PT include/exclude labels from git config.
+        # Use string.lower() since PT labels are case insensitive.
+        self._include_label = _gitflow.get('gitflow.pt.includelabel', None)
+        if self._include_label is not None:
+            self._include_label = self._include_label.lower()
+
+        self._exclude_labels = _gitflow.get('gitflow.pt.excludelables', None)
+        if self._exclude_labels is not None:
+            self._exclude_labels = self._exclude_labels.lower()
+            self._exclude_labels = self._exclude_labels.split(',')
+
         if _skip_story_download:
             return
         self._current_stories = list(_iter_current_stories())
@@ -235,6 +305,19 @@ class Release(object):
     def __iter__(self):
         for story in self._current_stories:
             if story.is_labeled('release-' + self._version):
+                if self._include_label:
+                    if story.is_labeled(self._include_label):
+                        # If the lable is defined, pick only the stories labeled
+                        # with that particular lable.
+                        yield story
+                    continue
+                if self._exclude_labels and \
+                        any([l for l in self._exclude_labels if story.is_labeled(l)]):
+                    # If the story contains any of the exclude labels, skip it.
+                    continue
+                # The include label is not defined and the story is not matching
+                # any of the exclude labels, if defined, so here we can just
+                # safely yield the story.
                 yield story
 
     def get_version(self):
@@ -245,32 +328,45 @@ class Release(object):
         for story in self.iter_candidates():
             story.assign_to_release(self._version)
 
-    def try_finish(self):
+    def try_stage(self):
         err = False
         for story in self:
+            if story.is_labeled('point me'):
+                err = True
+                print("    Story {0} labeled 'point me'!".format(story.get_url()))
+
+            label = None
+            for l in ('dupe', 'wontfix', 'cannot reproduce'):
+                if story.is_labeled(l):
+                    label = l
+                    print("    Story {0} labeled '{1}', skipping...".format(story.get_url(), l))
+                    break
+            if label is not None:
+                continue
+
             if not story.is_labeled('qa+'):
                 # Allow zero-point stories not to be QA'd since they are
                 # by definition refactoring stories.
                 if story.is_feature() and story.get_estimate() == 0:
                     continue
                 err = True
-                print "    Story not QA'd: " + story.get_url()
-            if story.is_labeled('point me'):
-                err = True
-                print "    Story labeled 'point me': " + story.get_url()
+                print("    Story {0} not QA'd!".format(story.get_url()))
         if err:
             raise StatusError("Pivotal Tracker check did not pass, operation canceled.")
 
-    def finish(self):
-        print 'Following stories were delivered for client acceptance as of release %s:' \
+    def stage(self):
+        print 'Following stories were staged for client acceptance as of release %s:' \
               % self._version
         for story in self:
-            story.deliver()
+            if not story.is_delivered() \
+                    and not story.is_accepted() \
+                    and not story.is_rejected():
+                story.deliver()
             sys.stdout.write('    ')
             story.dump()
 
-    def try_deliver(self):
-        self.try_finish()
+    def try_finish(self):
+        self.try_stage()
         err = False
         for story in self:
             if not story.is_accepted():
@@ -287,6 +383,13 @@ class Release(object):
 
     def iter_candidates(self):
         for story in self._current_stories:
+            if self._include_label and not story.is_labeled(self._include_label):
+                # Not labeled with the include label, skip it.
+                continue
+            if self._exclude_labels and \
+                    any([l for l in self._exclude_labels if story.is_labeled(l)]):
+                # Labeled with an exclude label, skip it.
+                continue
             if (story.is_feature() or story.is_bug()) \
                     and story.is_finished() \
                     and story.get_release() is None:
@@ -568,3 +671,44 @@ def show_release_summary(gitflow):
     #    if blockers:
     #        print "Story '%s' is blocked by stories: %s" % (story, blockers)
     #import ipdb; ipdb.set_tracce()
+
+
+def prompt_user_to_confirm_release(version):
+    release = Release(version)
+    any_assigned = False
+    print
+    print 'Stories already assigned to this release:'
+    for story in release:
+        sys.stdout.write('    ')
+        story.dump()
+        any_assigned = True
+    if not any_assigned:
+        print '    None'
+    print
+    any_candidate = False
+    any_pointme = False
+    print 'Stories to be newly assigned to this release:'
+    for story in release.iter_candidates():
+        if story.is_labeled('point me'):
+            sys.stdout.write('PM  ')
+            any_pointme = True
+        else:
+            sys.stdout.write('    ')
+        story.dump()
+        any_candidate = True
+    if not any_candidate:
+        print '    None'
+    print
+
+    if not any_candidate:
+        raise SystemExit('No new stories to be added to the release,' \
+                'aborting...')
+
+    if any_pointme:
+        raise PointMeError("Some stories are labeled with the 'point me' label")
+
+    if not release.prompt_for_confirmation():
+        raise SystemExit('Aborting...')
+
+def start_release(version):
+    Release(version).start()
